#pragma once

#include "ioh/experiment.hpp"
#include "ioh/common.hpp"
#include "ioh/logger.hpp" 
#include "ioh/problem.hpp"
#include "ioh/suite.hpp"

/** @defgroup Features Features
 * Classes you may use if you want to use IOH from your solver.
 * 
 */

/** @defgroup API API
 * What you may use to add a feature to IOH.
 * 
 */
